/*
maxmingap.cc

Use simulation to, for each adjusted margin of victory beta, compute max
and min probability of an adjusted m.o.v. >= beta, given that nA = nB.

Usage:
(requires boost library)
run: g++ maxmingap.cc -O2 -o maxmingap
run: ./maxmingap
In matlab, run the generated file maxminplot.m


Bo Waggoner

Created: 2011-07-26
Modified: 2012-04-15

This information is free.
*/


/*

Parameters:
-output file name: OUTFILE_NAME
-maximum votes castable by one person: maxv
-individual vote distribution: c[]
-number of alpha values: numalpha
-alpha values to use: alphalist[]
-number of trials at each n: trials
-random generator seed: seed
-whether to use exact or approximate method: USE_EXACT
    (if set to 0, approximate votes cast by a Gaussian; else,
     simulate each voter.)
-numbers of voters to use when simulating: n_min, n_step, n_max
-beta values to use when simulating: beta_min, beta_step, beta_max


Procedure:
for each value of alpha:
    for beta from beta_min to beta_max by beta_step:
        compute min and max probability of outcome >= beta over all n from n_min to m_max by n_step

print out the results in a matlab file

*/



#include <vector>
#include <fstream>
#include <cmath>
#include <ctime>
#include <boost/random/normal_distribution.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/variate_generator.hpp>


// output filename
const char* OUTFILE_NAME = "maxminplot.m";

// number of alphas to use and their values
const int numalpha = 3;
const double alphalist[] = {0.2, 0.5, 0.8};

// maximum number of votes to cast
const int maxv = 3;

// vote distribution c[i] = Pr[individual casts i votes]
const double c[maxv+1] = {0.1, 0.5, 0.2,0.2};

// properties of c[] to be computed at beginning of program
double c_cdf[maxv+1];
double mu;
double sigma;
double rho;

// number of trials to run at each n;
// probability = (number that exceed beta)/trials
const int trials = 10000;//1000000;

// random generator seed (set to -1 to use time(0))
int seed = 100;

// number of voters to use when simulating (note nA = nB = n/2)
const int n_min = 2;
const int n_step = 2;
const int n_max = 200;
int numn = ceil((n_max - n_min)/n_step) + 1;

// which beta values to use
const double beta_min = 0.2;
const double beta_step = 0.2;
const double beta_max = 3;
int numbeta = ceil((beta_max - beta_min)/beta_step) + 1;


const int USE_EXACT = 0;


// compute the cdf, mean, etc of c[]
void calcCurveProperties()
{
    double sum = 0;
    for(int i=0; i<=maxv; i++) sum += c[i] * i;
    mu = sum;
    sum = 0;
    for(int i=0; i<=maxv; i++) sum += (i - mu)*(i - mu) * c[i];
    sigma = sqrt(sum);
    sum = 0;
    for(int i=0; i<=maxv; i++)
    {
        double diff = (i-mu);
        diff *= diff*diff * c[i];
        sum += (diff > 0 ? diff : -diff);
    }
    rho = sum;
    c_cdf[0] = c[0];
    for(int i=1; i<=maxv; i++)
    {
        c_cdf[i] = c_cdf[i-1] + c[i];
    }

    //std::cout << mu << "  " << sigma << "  " << rho << "\n";
}

// inverse CDF method
// given a uniform [0,1] random variable, return the number of votes cast
// For large c[], this would be better implemented as a binary search,
// but for arrays of small size, the overhead isn't worth it
int getVotesCast(double cdfgenerated)
{
    int votescast = 0;
    while(votescast < maxv && cdfgenerated > c_cdf[votescast]) votescast++;
    return votescast;
}


// get the probability of exceeding beta when there are n voters, for each beta
// Approx: use gaussian approximation centered at mu*n with variance sigma^2 * n
//   for number of votes received by each alternative
void fillApproxProb(boost::mt19937 &randgen, int n, double alpha, double* prabove)
{
    for(int i=0; i<numbeta; i++) prabove[i] = 0;
    
    // votes for candidate j are approximately normal around mu*n/2, sigma*sqrt(n/2)
    boost::normal_distribution<> normA((double)n*mu/2.0f,sigma*sqrt((double)n/2.0f));
    boost::normal_distribution<> normB((double)n*mu/2.0f,sigma*sqrt((double)n/2.0f));
    
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<> > genA(randgen, normA);
    boost::variate_generator<boost::mt19937 &, boost::normal_distribution<> > genB(randgen, normB);
    
    for(int t=0; t<trials; t++)
    {
        double vA = genA();
        double vB = genB();
        
        double b = (vA - vB)/(pow(vA + vB,alpha));
        
        double diff = b - beta_min;
        if(diff < 0) continue;
        int index = ceil(diff/beta_step) + 1;
        for(int i=0; i<index && i<numbeta; i++)
        {
            prabove[i] += 1.0/(double)trials;
        }
    }
}

// get the probability of exceeding beta when there are n voters, for each beta
// Exact: simulate every voter
void fillExactProb(boost::mt19937 &randgen, int n, double alpha, double* prabove)
{
    for(int i=0; i<numbeta; i++) prabove[i] = 0;
    
    boost::uniform_real<> votercdf(0.0f,1.0f);
    boost::variate_generator<boost::mt19937 &, boost::uniform_real<> > votergencdf(randgen, votercdf);
    
    for(int t=0; t<trials; t++)
    {
        int vA = 0;
        for(int pA = 0; pA < n/2; pA++)
        {
            vA += getVotesCast(votergencdf());
        }
        int vB = 0;
        for(int pB = 0; pB < n/2; pB++)
        {
            vB += getVotesCast(votergencdf());
        }
        double b = ((double)(vA - vB))/(pow(vA + vB,alpha));
        
        double diff = b - beta_min;
        if(diff < 0) continue;
        int index = ceil(diff/beta_step) + 1;
        for(int i=0; i<index && i < numbeta; i++)
        {
            prabove[i] += 1.0/(double)trials;
        }
    }
}


int main()
{
    calcCurveProperties();
    
    // number of betas and ns to try
    numbeta = ceil((beta_max - beta_min)/beta_step) + 1;
    numn = ceil((n_max - n_min)/n_step) + 1;
    
    // maxps[i][j] = the maximum p-value for alpha=alphalist[i] at jth beta
    double* maxps[numalpha];
    double* minps[numalpha];
    for(int a=0; a<numalpha; a++)
    {
        maxps[a] = new double[numbeta];
        minps[a] = new double[numbeta];
    }
    double* temp_prabove = new double[numbeta];  // will track # margins >= beta
    
    // get a mersenne twister generator to use with the normal distributions
    if(seed < 0) seed = time(0);
    boost::mt19937 randgen(seed);

    for(int alphind=0; alphind<numalpha; alphind++)
    {
        for(int i=0; i<numbeta; i++)
        {
            maxps[alphind][i] = 0;
            minps[alphind][i] = 1;
        }
        
        int n = n_min;
        for(int i=0; i<numn; i++,n+=n_step)
        {
            if(USE_EXACT)
            {
                fillExactProb(randgen,n,alphalist[alphind],temp_prabove);
            }
            else
            {
                fillApproxProb(randgen,n,alphalist[alphind],temp_prabove);
            }
            
            for(int j=0; j<numbeta; j++)
            {
                if(temp_prabove[j] > maxps[alphind][j]) maxps[alphind][j] = temp_prabove[j];
                if(temp_prabove[j] < minps[alphind][j]) minps[alphind][j] = temp_prabove[j];
                
                // for these alphas, we've proven that there is a large enough n where this is true
                // to see what bound is actually found by the simulations, comment these lines out
                if(alphalist[alphind] < 0.5) maxps[alphind][j] = 0.5;
                if(alphalist[alphind] > 0.5) minps[alphind][j] = 0;
            }
        }
    }

    // save it all to file
    std::ofstream file;
    file.open(OUTFILE_NAME);
    
    file << "\% " << OUTFILE_NAME << "\n";
    file << "\%\n\% Plotting the chances of getting the wrong 'outcome'\n\% as measured with certain alphas.\n";
    file << "\%\n\n\% Bo Waggoner\n\% Created: ";
    time_t rawtime;
    time(&rawtime);
    file << asctime(localtime(&rawtime)) << "\n%\n\n";
    
    file << "close all;\nclear all;\n\n";
    
    file << "mu = " << mu << ";\nsigma = " << sigma << ";\n\n";
    
    file << "n_min = " << n_min << ";\nn_step = " << n_step << ";\nn_max = " << n_max << ";\n";
    file << "ns = n_min:n_step:n_max;\n\n";
    
    file << "numalpha = " << numalpha << ";\n";
    file << "alphalist = [";
    for(int i=0; i<numalpha; i++)
    {
        file << alphalist[i];
        if(i < numalpha-1) file << ",";
    }
    file << "];\n\n";
    
    file << "beta = [";
    for(double b=beta_min; b<=beta_max; b+=beta_step)
    {
        file << b;
        if(b < beta_max) file << ",";
    }
    file << "];\n\n";


    file << "maxps = [";
    for(int alphind=0; alphind<numalpha; alphind++)
    {
        int step=0;
        for(double b=beta_min; b<=beta_max; b+=beta_step)
        {
            file << maxps[alphind][step];
            if(b < beta_max) file << ",";
            step++;
        }
        if(alphind < numalpha-1) file << ";";
    }
    file << "];\n\n";
    
    file << "minps = [";
    for(int alphind=0; alphind<numalpha; alphind++)
    {
        int step=0;
        for(double b=beta_min; b<=beta_max; b+=beta_step)
        {
            file << minps[alphind][step];
            if(b < beta_max) file << ",";
            step++;
        }
        if(alphind < numalpha-1) file << ";";
    }
    file << "];\n\n";

    file << "colors = {'k','b','r','g','m'};";
    file << "linesty = {'-*';'-o';'-+';'-.';'-s';'-^'};\n";
    file << "skip = 3;\n\n";

    file << "figure(1);\nfor i=1:numalpha,\n    hnew = plot(beta(1:skip:end),maxps(i,1:skip:end),'-*','color',colors{i});\n    hold on;\n";
    file << "    legstr = strcat('maximum, \\alpha = ',num2str(alphalist(i)));\n";
    file << "    if i==1,\n        legend(legstr);\n    else\n";
    file << "        [LEGH,OBJH,OUTH,OUTM] = legend;\n        legend([OUTH;hnew],OUTM{:},legstr);\n";
    file << "    end\nend\n\n";

    file << "for i=1:numalpha,\n    hnew = plot(beta(1:skip:end),minps(i,1:skip:end),'-o','color',colors{i});\n    hold on;\n";
    file << "    legstr = strcat('minimum, \\alpha = ',num2str(alphalist(i)));\n";
    file << "    [LEGH,OBJH,OUTH,OUTM] = legend;\n    legend([OUTH;hnew],OUTM{:},legstr);\n";
    file << "    end\naxis([0 beta(end) 0 1]);\n\n";
    file << "xlabel('\\beta');\n";
    file << "ylabel('max Pr[margin of victory \\geq \\beta]');";

    for(int i=0; i<numalpha; i++) {delete[] maxps[i]; delete[] minps[i];}
    delete[] temp_prabove;
}



